! Module: imagesDal.f90 
! Author: P.Videler
! Version: 0.5
! 
! (c) 1998 European Space Agency 
!
! Description: module for reading/writing a set of related images 
!               (image, badpix, weight) to a DAL dataset
! 
! Comments:  Having image, badpix and/or weights file in the same datasets
!            results in multiple copies of the dataset being opened by the DAL
! 
! Modifications:
!   
! 
! $Id: imagesdal.f90,v 1.6 2007/06/15 15:48:03 rsaxton Exp $
! 
module imagesDal

  use saslib
  use obj_images
  use dal
  use param

  implicit none 

  !......................................................................
  private    ! change default to private
  public  :: &
       imagesDal_t,  &        
       init,         &
       free,         &
       getImages,    &
       createImages,        &
       addImage2eventTable, &
       addTable2Image
  !......................................................................

  type imagesDal_t
     type(images_t)       :: images
     type(DataSetT)       :: imageSet, badpixSet, weightSet
  endtype imagesDal_t

  interface init
     module procedure imagesDal_init, imagesDal_init_vector
  end interface

  interface free
     module procedure imagesDal_free, imagesDal_free_vector
  end interface

contains

  ! -------------------------------------------------------------------------
  logical function presentAndNE(optionalStr, value) result(bool)
    ! ----------------------------------------------------------------------- 
    character(len=*),    intent(in), optional  :: optionalStr
    character(len=*),    intent(in)            :: value

    !   Utility function to replace the illegal f90 construct below. SPR
    !!     if (present(optinalArg) .AND. optionalArg == value)  then 

    bool = .false.
    if (present(optionalStr)) bool = (optionalStr /= value)

  end function presentAndNE


  !---------------------------------------------------------------------------
  subroutine imagesDal_init(aa)
    !---------------------------------------------------------------------------
    type(imagesDal_t), intent(out)  :: aa

    call init(aa%images)

  end subroutine imagesDal_init

  !---------------------------------------------------------------------------
  subroutine imagesDal_init_vector(aa)
    !---------------------------------------------------------------------------
    type(imagesDal_t), intent(out)  :: aa(:)
    integer                         :: i  

    do i=1, size(aa)
       call init(aa(i)%images)
    enddo

  end subroutine imagesDal_init_vector

  !---------------------------------------------------------------------------
  subroutine imagesDal_free(aa)
    !---------------------------------------------------------------------------
    type(imagesDal_t), intent(inout)  :: aa
    character(len=*),   parameter  :: fn='imagesDal_free : '

    if (doImage (aa%images)) then
       write(errstr,*) fn//'doing release on imageSet ', name(aa%imageSet)
       call message(appLibMsg, Verbose)
       call release(aa%imageSet)
    endif

    ! ##################################
    ! release on other dataset temporarily disabled
    ! ##################################

!!$  if (doWeight(aa%images)) then
!!$    write(errstr,*) fn//'doing release on weightSet ', name(aa%weightSet)
!!$        call message(appLibMsg, Verbose)
!!$    call release(aa%weightSet)
!!$  endif
!!$
!!$  if (doBadpix(aa%images)) then
!!$    write(errstr,*) fn//'doing release on badpixSet ', name(aa%badpixSet)
!!$    call message(appLibMsg, Verbose)

!!$    call release(aa%badpixSet)
!!$  endif

    call free(aa%images)

  end subroutine imagesDal_free

  !---------------------------------------------------------------------------
  subroutine imagesDal_free_vector(aa)
    !---------------------------------------------------------------------------
    type(imagesDal_t), intent(inout)  :: aa(:)
    character(len=*),   parameter     :: fn='imagesDal_free_vector : '
    integer                           :: i  

    do i=1, size(aa)
       call  imagesDal_free(aa(i))
    enddo

  end subroutine imagesDal_free_vector

  ! ====================================================================
  function GetImages(aa, imageSetName, imageName, badpixSetName, &
       badpixName, weightSetName, weightName) result(xx) 
    ! ====================================================================
    type(imagesDal_t),   intent(inout), target :: aa
    type(images_t),      pointer               :: xx
    character(len=*),    intent(in)            :: imageSetName
    character(len=*),    intent(in), optional  :: imageName
    character(len=*),    intent(in), optional  :: badpixSetName, badpixName
    character(len=*),    intent(in), optional  :: weightSetName, weightName

    type(ArrayT)                   :: dalImage, dalBadpix, dalWeight
    integer                        :: lowl(2), upr(2)
    character(len=*),   parameter  :: fn='GetImages : '

    ! ----------------------
    ! Get image from DAL
    ! ----------------------
    
    xx => aa%images     ! alias

    aa%imageSet = dataSet(imageSetname, read)

    if (presentAndNE(imageName, 'PRIMARY')) then
       dalImage = array(aa%imageSet, imageName, read)
    else
       write(errstr,*) fn// 'numberOfBlocks= ', numberOfBlocks(aa%imageset)
       call message(appLibMsg, Verbose)

       dalImage = array(aa%imageSet, 0, read)          ! first image
       call message(appLibMsg, Verbose, fn//'Read FIRST image from data set.')
    endif

    write(errstr,*) fn//'Image "', trim(name(dalImage)),  &
         '" of dimensions ', dimensions(dalImage),          &
         ' read from dataset "', trim(name(aa%imageSet)), '"'
       call message(appLibMsg, Verbose)

!!$ ! this is how sub image could be selected once it gets implemneted in the DAL
!!$  call arraySeek(dalImage, (/ 5,5/),(/ 7,7/))

    xx = aa%images     ! alias

    xx%image => real32Array2Data(dalImage)
       call message(appLibMsg, Verbose, fn//'Image data interpreted as real32.')

    xx%lowleft = lbound(xx%image)

    ! write(*,*) fn//'image dimenions (DAL) : ', dimensions(dalImage)
    ! write(*,*) fn//'shape image (F90)     : ', shape(xx%image)

    ! Extra check on image dimensions
    if (ANY(dimensions(dalImage) /= shape(xx%image)))  &
         call error (fn ,'Something wrong with the DAL image dimensions.')


    ! ---------------
    !  Badpixel map
    ! ---------------

    ! defined as:  badpix=true  (/=0) (pixel is bad) 
    !              badpix=false (0) (pixel is good)

    if (booleanParameter('usebadpix')) then

!!$     Current DAL (v1.0) causes a copy of the dataset to be loaded
!!$     in memory when imageSet and badpixSet are the same.

       if (presentAndNE(badpixSetName, 'default')) then
          aa%badpixSet = dataSet(badpixSetname, read)
       else
          ! aa%badpixSet = dataSet(imageSetname, read) 
          aa%badpixset = aa%imageset
       endif


       ! default badpixname = imageName + '_badpix'

       if (presentAndNE(badpixName, 'default')) then
          dalBadpix = array(aa%badpixSet, badpixName, read)
       else
          dalBadpix = array(aa%badpixSet, trim(name(DalImage))//'_badpix', read)   
       endif

       write(errstr,*) fn//'BadpixImage "', trim(name(dalBadpix)),  &
            '" of dimensions ', dimensions(dalBadpix),          &
            ' read from dataset "', trim(name(aa%badpixSet)), '"'
       call message(appLibMsg, Verbose)

       xx%badpix => int32Array2Data(dalBadpix)
       call message(appLibMsg, Verbose,fn//'Badpix data interpreted as 32 bit integer.' )

    endif

    ! ---------------
    !  Error map
    ! ---------------

    if (booleanParameter('useweight')) then

       if (presentAndNE(weightSetName, 'default')) then
          aa%weightSet = dataSet(weightSetname, read)
       else
          ! aa%weightSet = dataSet(imageSetname, read) 
          aa%weightSet = aa%imageset
       endif

       if (presentAndNE(weightName, 'default')) then
          dalWeight = array(aa%weightSet, weightName, read)
       else
          dalWeight = array(aa%weightSet, trim(name(DalImage))//'_weight', read)
       endif

       write(errstr,*) fn//'WeightImage "', trim(name(dalWeight)),  &
            '" of dimensions ', dimensions(dalWeight),          &
            ' read from dataset "', trim(name(aa%weightSet)), '"'
       call message(appLibMsg, Verbose)

       xx%weight => real32Array2Data(dalWeight)
       call message(appLibMsg, Verbose, fn//'Weight data interpreted as real32.' )

    endif

    ! ---------------
    !  check Data
    ! ---------------

    call check(xx)

    ! ---------------
    !  set window (optional)
    ! ---------------

    if (booleanParameter('window')) then
       lowl(1) = intParameter('xlowleft')  
       lowl(2) = intParameter('ylowleft')  
       upr(1)  = intParameter('xupperright')  
       upr(2)  = intParameter('yupperright')  

       call imageWindow(xx, lowleft=lowl, upright=upr)

    endif

  end function GetImages

  ! ====================================================================
  function createImages(aa, imsize, imageSetName, imageName, badpixName, &
       weightName) result(xx) 
    ! ====================================================================

    ! note: image-, badpix- and weight- images are all created in a single dataset

    type(imagesDal_t),   intent(inout), target :: aa
    type(images_t),      pointer               :: xx
    integer            , intent(in)            :: imsize(2)
    character(len=*),    intent(in)            :: imageSetName
    character(len=*),    intent(in), optional  :: imageName, badpixName, weightName  

    type(ArrayT)                   :: dalImage, dalBadpix, dalWeight
    character(len=*),   parameter  :: fn='createImages : '

    xx => aa%images     ! alias

    ! -------------------------------
    ! Create image array  (using DAL)
    ! -------------------------------

    
    aa%imageSet = dataSet(imageSetname, create)

    print*,' Created dataset: ',imageSetname

    if (presentAndNE(imageName, 'default')) then
       dalImage = addArray(aa%imageSet, imageName, real32, imsize)
    else
       dalImage = addArray(aa%imageSet, 'theImage' , real32, imsize)
    endif

    write(errstr,*) fn//'Image "', trim(name(dalImage)),  &
         '" of dimensions ', dimensions(dalImage),          &
         ' added to newly created dataset "', trim(name(aa%imageSet)), '"'
    call message(appLibMsg, Verbose)

    xx = aa%images     ! alias

    xx%image => real32Array2Data(dalImage)
    call message(appLibMsg, Verbose, fn//'Image data interpreted as real32.'  )

    xx%image=0
    call message(appLibMsg, Verbose,fn//'All pixels initialized to 0.' )

    xx%lowleft = lbound(xx%image)


    ! ---------------------------------
    !  Create array with Badpixel data
    ! ---------------------------------

    if (booleanParameter('usebadpix')) then

       ! default badpixname = imageName + '_badpix'
       if (presentAndNE(badpixName, 'default')) then
          dalBadpix = addArray(aa%imageSet, badpixName, integer32, imsize)
       else
          dalBadpix = addArray(aa%imageSet, trim(name(DalImage))//'_badpix', integer32, imsize)   
       endif

       write(errstr,*) fn//'BadpixImage "', trim(name(dalBadpix)),  &
            '" of dimensions ', dimensions(dalBadpix),          &
            ' added to dataset "', trim(name(aa%imageSet)), '"'
       call message(appLibMsg, Verbose)

       xx%badpix => int32Array2Data(dalBadpix)
       call message(appLibMsg, Verbose, fn//'Badpix data interpreted as 32 bit integer.')

       xx%badpix=0
       call message(appLibMsg, Verbose, fn//'All pixels initialized to good (0).')

    endif

    ! ---------------------------------
    !  Create array with weights
    ! --------------------------------

    if (booleanParameter('useweight')) then

       if (presentAndNE(weightName, 'default')) then
          dalWeight = addArray(aa%imageSet, weightName, real32, imsize)
       else
          dalWeight = addArray(aa%imageSet, trim(name(DalImage))//'_weight', real32, imsize)
       endif

       write(errstr,*) fn//'Weights Image "', trim(name(dalWeight)),  &
            '" of dimensions ', dimensions(dalWeight),          &
            ' added to dataset "', trim(name(aa%imageSet)), '"'
       call message(appLibMsg, Verbose)

       xx%weight => real32Array2Data(dalWeight)
        call message(appLibMsg, Verbose, fn//'Weight data interpreted as real32.')

       xx%weight=0
       call message(appLibMsg, Verbose, fn//'All weights initialized to 0.' )

    endif

    ! ---------------
    !  check Data
    ! ---------------

    call check(xx)

  end function createImages

  ! ====================================================================
  function addImage2eventTable(xx, set, tblname) result(tbl)
    ! ====================================================================

    type(images_t) ,     intent(in)    :: xx
    type(DataSetT),      intent(in)    :: set
    character(len=*),    intent(in)    :: tblname
    type(TableT)                       :: tbl

    integer,             parameter     :: ncol=5
    type(ColumnT)                      :: col(ncol)
    integer(kind=int16),  pointer      :: xpos(:), ypos(:)
    integer(kind=int32),  pointer      :: flag(:)
    real(single),         pointer      :: cont(:), weight(:)
    integer                            :: nrows, sz(2)
    integer                            :: lb(2), ub(2), shift(2)  
    integer                            :: i, j, n

    integer,              parameter    :: pixellowleft = 0
    character(len=*),     parameter    :: fn='addImage2eventTable : '
    logical                            :: doweight_ = .false.

    ! checks
    if (.not. doImage(xx)) call error (fn ,' Image not defined')

    ! use lowerbounds for images
    lb  = lbound(xx%image)  
    ub  = ubound(xx%image)

    nrows = size(xx%image)

    write (errstr,*) "Image bounds ",lb,ub,"nrows",nrows
    call message(AppMsg,Noisy)

    ! exclude bad pixels from event table. DISABLED PV
    !!  if (doBadpix(xx)) nrows = nrows - count(xx%badpix /= 0 )

    ! add table 
    tbl = addTable(set, tblname, nrows)
    write(errstr,*) fn//' creating table with ', nrows,' rows.'
       call message(appLibMsg, Verbose)

    ! define columns
    col(1) = addColumn(tbl, stringParameter("xposcolumn") , integer16, '#')
    col(2) = addColumn(tbl, stringParameter("yposcolumn"),  integer16, '#')
    col(3) = addColumn(tbl, stringParameter("countscolumn"),   real32, '#')
    xpos =>  int16Data(col(1))
    ypos =>  int16Data(col(2))
    cont => real32Data(col(3))

    if (doWeight(xx)) then
       col(4) = addColumn(tbl, stringParameter("weightcolumn"),   real32, '#')
       weight => real32Data(col(4))
       doweight_ = .true.
    endif

    if (doBadpix(xx)) then
       col(5) = addColumn(tbl, stringParameter("flagcolumn"),   integer32, '#')
       flag => int32Data(col(5))
    endif

    ! images start at lb. SAS convension index start at pixellowleft 
    shift = pixellowleft - lb

    sz(1) = size(xx%image, dim=1) 
    sz(2) = size(xx%image, dim=2)

    ! must fit in int16 xpos ypos array's
    if (ANY(sz > 32767))            call error (fn ,' Image dimension exceeds 16bit integer (2^15-1)')
    if (ANY(ABS(sz+shift) > 32767)) call error (fn ,' Pixel coordinates  exceeds 16bit integer (2^15-1)')

    ! badpix in separate loop for better performance
    if (doBadpix(xx)) then 
       n = 0
       do i=lb(1), ub(1)
          do j = lb(2), ub(2)  
             ! if (xx%badpix(i,j) /= 0) cycle      ! skip bad points
             flag(n) = xx%badpix(i,j)
             xpos(n) = i + shift(1)
             ypos(n) = j + shift(2)
             cont(n) = xx%image(i,j)
             if (doweight_) weight(n) = xx%weight(i,j)
             n = n + 1
          enddo
       enddo

    else

       n = 0
       do i=lb(1), ub(1)               ! loops could be avoided using vector functions
          do j = lb(2), ub(2)          ! easier/clearer with do loop?  
             xpos(n) = i + shift(1)
             ypos(n) = j + shift(2)
             cont(n) = xx%image(i,j)
             if (doweight_) weight(n) = xx%weight(i,j)
             n = n + 1
          enddo
       enddo

    endif

    ! release colums. Note: table must be released in calling routine.
    do i=1, 3
       call release(col(i))
    enddo
    if (doweight(xx)) call release(col(4))
    if (dobadpix(xx)) call release(col(5))

  end function addImage2eventTable

  ! ====================================================================
  subroutine addTable2Image(tbl, xx) 
    ! ====================================================================
    !  use histogram

    type(TableT),        intent(in)    :: tbl
    type(images_t),      pointer       :: xx

    integer(kind=int16),  pointer      :: xpos(:), ypos(:)
    integer(kind=int32),  pointer      :: flag(:)
    real(kind=single),    pointer      :: counts(:), weight(:)
    logical                            :: doflag=.false.
    logical                            :: doweight=.false.
    character(len=*),    parameter     :: fn='addTable2Image : '
    !  real(single)                       :: min(2), max(2)
    integer                             :: n, i,j, shift(2)

    ! type(histogram2d_t),  pointer      :: im, w, b

    ! checks
    if (.not. doImage(xx)) call error (fn ,' Image not defined')

    ! map columns
    xpos   =>  int16Data( column(tbl, stringParameter("xposcolumn")  , read))
    ypos   =>  int16Data( column(tbl, stringParameter("yposcolumn")  , read))
    counts => real32Data( column(tbl, stringParameter("countscolumn"), read))

    if (hasColumn(tbl, trim(stringParameter("flagcolumn")))) doflag = .true.

    if (doflag) then
       flag => int32Data( column(tbl, stringParameter("flagcolumn"), read))
    endif

    if (hasColumn(tbl, trim(stringParameter("weightcolumn")))) doweight = .true.

    if (doweight) then
       weight => real32Data( column(tbl, stringParameter("weightcolumn"), read))
    endif

    ! -------------------
    ! bin into image (this should be replaced with histolib binning)
    ! -------------------

    ! images start at lb. SAS convension index start at pixel lowleft

    shift = lbound(xx%image) - xx%lowleft

    ! processing loop for all events
    do n = 0, numberOfRows(tbl) - 1 

       ! binsize 1 (trivial) ######
       i = xpos(n) + shift(1)
       j = ypos(n) + shift(2)

       if (doflag) then
          xx%badpix(i,j) = flag(n)
          ! if (flag(n) /= 0) cycle      ! skip flagged pixels
       endif

       xx%image(i,j) = xx%image(i,j) + counts(n) 

       if (doweight) xx%weight(i,j) = xx%weight(i,j) + weight(n)

    enddo

    ! -------------------
    ! binning into xx%image using histolib
    ! -------------------

!!$  call initHisto()
!!$
!!$  nbins = shape(xx%image)
!!$  min = REAL(xx%lowleft, kind=single)
!!$  max = min + nbins
!!$
!!$  ! setup histogram
!!$  im => newHisto(name='image', nbins=nbins, min=min , max=max, hArray= xx%image )
!!$
!!$  ! bin data (xpos ypos converted to real).
!!$  if (doflag) then
!!$     call addHisto(im, REAL(xpos), REAL(ypos), weight=counts, mask=flag)
!!$  else 
!!$     call addHisto(im, REAL(xpos), REAL(ypos), weight=counts)
!!$  endif
!!$
!!$  if (dbg) write(*,*) fn// ' debug histogram'
!!$  call report(im)
!!$
!!$  ! same for weights
!!$  if (doweight) then
!!$     w => newHisto(name='weight', nbins=nbins, min=min , max=max, hArray= xx%weight )
!!$
!!$     if (doflag) then
!!$        call addHisto(w, REAL(xpos), REAL(ypos), weight=weight, mask=flag)
!!$     else
!!$        call addHisto(w, REAL(xpos), REAL(ypos), weight=weight)
!!$     endif
!!$
!!$     call report(w)
!!$  endif
!!$
!!$!!! ####################
!!$!!! THis does not work because hArray must be real and xx%badpix is integer
!!$
!!$!!  ! same for flag (badpix)
!!$!!  if (doflag) then
!!$!!     b => newHisto(name='badpix', nbins=nbins, min=min , max=max, hArray= xx%badpix )
!!$!!
!!$!!     call addHisto(b, REAL(xpos), REAL(ypos), weight=REAL(flag))
!!$!!
!!$!!     call report(b)
!!$!!  endif
!!$!!!  ####################
!!$
!!$  ! free histograms
!!$  call freeHisto()

  end  subroutine addTable2Image

end module imagesDal


