module obj_images

 use saslib
 implicit none 

 !......................................................................
 private    ! change default to private
 public  ::      &
      images_t,     &        
      cLen,         &
      init,         &
      free,         &
      doBadpix,     &
      doWeight,     &
      doImage,      &
      imageWindow,  &
      imageScale,   &
      check
 !......................................................................

integer, parameter :: cLen=256    ! length of name strings

!!$#########################################################################
!!$
!!$  Badpix changed from boolean into int32 because the DAL/FITS does not 
!!$  support the bool type for arrays.
!!$
!!$#########################################################################

 type images_t
    real(single),     pointer  :: image(:,:)
    real(single),     pointer  :: weight(:,:)
!!    logical,          pointer  :: badpix(:,:)  
    integer(int32),      pointer  :: badpix(:,:)  

    integer                  :: lowleft(2)        ! lower left corner
    integer                  :: scalefactor       ! used for image blowup
 endtype images_t

! note: lowleft is NOT the same as LBOUND(image).
!       lowleft gives the pixel numbers of the original image where LBOUND(image) 
!       currently correspond to. After windowing lowleft differs from LBOUND(image).

 interface init
    module procedure imageInit
 end interface

 interface free
    module procedure imageFree
 end interface

 interface check
    module procedure checkShape
 end interface

contains

 !---------------------------------------------------------------------------
 subroutine imageInit(xx)
  !---------------------------------------------------------------------------
  type(images_t), intent(out)  :: xx
  character(len=*),    parameter   :: fn='obj_images: imageInit : '

  call message(appLibMsg, Verbose, fn//' initializing images')
  nullify(xx%image)
  nullify(xx%weight)
  nullify(xx%badpix)
  xx%lowleft = (/0,0/)
  xx%scalefactor = 1_single

 end subroutine imageInit

 !---------------------------------------------------------------------------
 subroutine imageFree(xx)
  !---------------------------------------------------------------------------
  type(images_t), intent(inout)  :: xx
  character(len=*),    parameter   :: fn='obj_images: imageFree : '

  ! write(*,*) fn//' free images'

  ! Can't deallocate if allocated by DAL
  ! only deallocate if allocated by imageScale ie scale /= 1
  if (xx%scalefactor /= 1) then
     call message(appLibMsg, Verbose, fn//' deallocating images')
     if (doImage(xx))  deallocate(xx%image)
     if (doBadpix(xx)) deallocate(xx%badpix)
     if (doWeight(xx)) deallocate(xx%weight)
  endif

  call init(xx)

 end subroutine imageFree

 !---------------------------------------------------------------------------
 logical function doImage(xx)
  !---------------------------------------------------------------------------
  type(images_t), intent(in)  :: xx
  doImage=.false.
  if (associated(xx%image)) doImage = .true.  
 end function doImage

 !---------------------------------------------------------------------------
 logical function doBadpix(xx)
  !---------------------------------------------------------------------------
  type(images_t), intent(in)  :: xx
  doBadpix=.false.
  if (associated(xx%badpix)) doBadpix = .true.  
 end function doBadpix

 !---------------------------------------------------------------------------
 logical function doWeight(xx)
  !---------------------------------------------------------------------------
  type(images_t), intent(in)  :: xx
  doWeight=.false.
  if (associated(xx%Weight)) doWeight = .true.  
 end function doWeight

 !---------------------------------------------------------------------------
 subroutine imageWindow(xx, lowleft, upright)
  !---------------------------------------------------------------------------
  type(images_t),   intent(inout) :: xx
  integer,          intent(in)    :: lowleft(2), upright(2)
 
  integer                         :: ll(2), ur(2), lb(2), ub(2)
  character(len=*),   parameter   :: fn='obj_images: imageZoom : '

  ll = lowleft ;  ur = upright

  ! make sure corners do not fall outside edge image
  lb = lbound(xx%image) 
  ub = ubound(xx%image)
  if (ll(1) < lb(1) .or. ll(1) > ub(1)) ll(1) = lb(1)
  if (ll(2) < lb(2) .or. ll(2) > ub(2)) ll(2) = lb(2)
  if (ur(1) > ub(1) .or. ur(1) < lb(1)) ur(1) = ub(1)
  if (ur(2) > ub(2) .or. ur(2) < lb(2)) ur(2) = ub(2)

  if (ANY (ll == lb) .or. ANY(ur == ub)) call message (appLibMsg, sparse, &
      fn//' Image window limits restricted by original imagebounds' ) 

  write(errstr,*) fn//' lowleft upright=', ll, ur
  call message(appLibMsg, Verbose)
  xx%image => xx%image(ll(1):ur(1) ,ll(2):ur(2) )

  ! same window for badpix and weight
  if (doBadpix(xx)) xx%badpix  => xx%badpix( ll(1):ur(1) ,ll(2):ur(2) )
  if (doWeight(xx)) xx%weight  => xx%weight( ll(1):ur(1) ,ll(2):ur(2) )

  ! save original low left corner
  xx%lowleft = xx%lowleft + ll

  ! check image for zero size etc.
  call check(xx)

 end subroutine imageWindow

 !---------------------------------------------------------------------------
 subroutine imageScale(xx, scaleFactor)
  !---------------------------------------------------------------------------
  type(images_t), intent(inout)  :: xx
  integer,        intent(in)     :: scaleFactor 

  integer,      dimension(2) :: lo, hi, lb, ub, sz
  integer                    :: n, i, j

  real(single),       pointer  :: oldimage(:,:), oldweight(:,:)
  !! logical,            pointer  :: oldbadpix(:,:)
  integer(kind=int32), pointer  :: oldbadpix(:,:)

  logical                    :: badpix, weight, image

  character(len=*),    parameter   :: fn='obj_images: imageScale : '

  n = scaleFactor
  if (n<1) RETURN          ! only n>=1 , n=1 usefull for testing only

  write(errstr,*) fn//' Zooming with scalefactor of ',n
  call message(appLibMsg, Verbose)

  xx%scalefactor = n* xx%scalefactor

  ! use logicals for shorter notation
  image  = doImage(xx)
  badpix = doBadPix(xx)
  weight = doWeight(xx)

  ! set pointers to original images
  oldimage  => xx%image
  oldweight => xx%weight
  oldbadpix => xx%badpix
   
  ! array bounds and sizes
  lb = lbound(xx%image)
  ub = ubound(xx%image)
  sz(1) = size(xx%image, dim=1)
  sz(2) = size(xx%image, dim=2)

  ! allocate n times larger images
  lo = lb                        ! same lowerbounds as original image
  hi = lb - 1 + n*sz             ! upper bound big image 
  write(errstr,*) fn//'allocating big images of size:',  lo,' to ', hi
  call message(appLibMsg, Verbose)

  if (image)  allocate( xx%image ( lo(1):hi(1), lo(2):hi(2) ))
  if (badpix) allocate( xx%badpix( lo(1):hi(1), lo(2):hi(2) ))
  if (weight) allocate( xx%weight( lo(1):hi(1), lo(2):hi(2) ))

  ! fill big images.
  do i=lb(1), ub(1)
     do j=lb(2), ub(2)
        lo(1) = lb(1) + n*(i-lb(1))    !  ?? yep 
        lo(2) = lb(2) + n*(j-lb(2))    !  
        hi = lo + n -1                 ! each original pixel is repeated n times
        if (image)  xx%image ( lo(1):hi(1) , lo(2):hi(2)) = oldimage (i,j)
        if (badpix) xx%badpix( lo(1):hi(1) , lo(2):hi(2)) = oldbadpix(i,j)
        if (weight) xx%weight( lo(1):hi(1) , lo(2):hi(2)) = oldweight(i,j)
     enddo
  enddo

 end subroutine imageScale

  !---------------------------------------------------------------------------
 subroutine checkShape(xx)
  !---------------------------------------------------------------------------
  type(images_t),      intent(in)  :: xx
  character(len=*),    parameter   :: fn='obj_images: checkShape : '
  integer                          :: lb(2), ub(2)

  if (size(xx%image) < 1) then
     write(errstr,*) fn//': Size image < 1. Image size = ',size(xx%image)
     call error (fn)
  endif

  lb = lbound(xx%image) ; ub = ubound(xx%image)
  write(errstr,*) fn//' image bounds are:', lb, ub
  call message(appLibMsg, Verbose)

  ! both sizes and bounds  of image, badpix and weight must match

  if (doBadpix(xx)) then
     if (ANY(lbound(xx%badpix) /= lb)) then
        write(errstr,*) fn//': Lower bounds bad pixel array do not match image array. &
             & Image: ',lb,' :Badpixel ',lbound(xx%badpix)
        call error (fn)
     endif

     if (ANY(shape(xx%image) /= shape(xx%badpix)))  then 
        write(errstr,*) fn//': Dimensions bad pixel array do not match image array. &
             & Shape Image: ',shape(xx%image),' :Shape Badpixel ',shape(xx%badpix)
        call error (fn)
     endif
  endif

  if (doWeight(xx)) then
     if (ANY(lbound(xx%weight) /= lb)) then
        write(errstr,*) fn//': Lower bounds weight array do not match image array. &
             & Image: ',lb,' :Weight ',lbound(xx%weight)
        call error (fn)
     endif

     if (ANY(shape(xx%image) /= shape(xx%weight)))  then 
        write(errstr,*) fn//': Dimensions weights array do not match image array. &
             & Shape Image: ',shape(xx%image),' :Shape weights',shape(xx%weight)
        call error (fn)
     endif
  endif

 end subroutine checkShape

end module obj_images
