! Module: tableStat.f90 
! Author: P.Videler
! Version: 0.9
! 
! (c) 1997 European Space Agency 
!       Part of the XMM Science Analysis System.
!
! Description: image statistics on eventlist based images.
! 
! Comments: Made into separate module because of globals needed 
!           by 'forEachSubTable'. 
! 
! Modifications: 
! 
! $Id: tablestat.f90,v 1.4 2000/05/17 11:49:49 pvideler Exp $
! 
module tableStat
 use saslib
 use dal
 use obj_imageStat
 use param

 implicit none 

 !......................................................................
 private    ! change default to private
 public  :: imageStat_table
 !......................................................................

 ! module variables
 type(imageStat_t)    :: s          ! statistics
 logical              :: doflag=.false.

contains 

 ! ====================================================================
 function imageStat_table (tbl)
  ! ====================================================================
  type(TableT),    intent(in)  :: tbl
  type(imageStat_t)            :: imageStat_table
  integer                      :: n

  ! init stat results type
  call init(s)

  ! check for flag column
  if (hasColumn(tbl, trim(stringParameter("flagcolumn")))) doflag = .true.

  ! loop over table using iterator    
  call forEachSubTable(tbl, processSubTable)
 
  ! center
  if (s%total == 0) then
     call warning ('imagelib/src/tablestat','Sum of image counts is zero. &
                    & Can not derive center position for this image. &
                    & Using default (-1,-1) for center')
  else 
    s%center = s%sigma / s%total
  endif

  s%ngoodpix = numberOfRows(tbl) - s%nbadpix

 ! variance
  n = s%ngoodpix
  if (n>0) then
    s%variance = (s%sumOfSquares - s%total**2 /n) /n
  else
   s%variance = 0
  endif 

  imageStat_table = s

 end function imageStat_table

 ! ====================================================================
 subroutine processSubTable(tbl)
  ! ====================================================================
  ! note: 'forEachSubTable' allows this subroutine to have only one argument (table)

  type(TableT),           intent(in) :: tbl
  integer(kind=int16),  pointer      :: xpos(:), ypos(:)
  integer(kind=int32),  pointer      :: flag(:)
  real(kind=single),    pointer      :: counts(:)
  integer  :: i
  type(columnT) :: col(4)

  ! map columns
  
  col(1) = column(tbl, stringParameter("xposcolumn"), read)
  col(2) = column(tbl, stringParameter("yposcolumn"), read)
  col(3) = column(tbl, stringParameter("countscolumn"), read)
  xpos   =>  int16Data(col(1))
  ypos   =>  int16Data(col(2))
  counts => real32Data(col(3))

  if (doflag) then
     col(4) = column(tbl, stringParameter("flagcolumn"), read)
     flag => int32Data(col(4))
  endif

  ! process table
  ! (using doloop rather then array intrinsics because of sigma) 

!!$  write(*,*) 'debug  nrows', numberOfRows(tbl)
!!$  write(*,*) 'debug  count', count(tbl)
!!$  write(*,*) 'debug size', size(xpos)

  ! do i= 0, numberOfRows(tbl) - 1  ! ### does not work in lowmem processSubtable
   do i= 0, count(tbl) - 1  
  ! do i= 0, size(xpos) - 1         ! alternative to count

     if (doflag) then
        if (flag(i) /= 0) then                ! flaged is non zero
           s%nbadpix = s%nbadpix + 1
           cycle                              ! skip flaged  pixels
        endif
     endif

     s%total = s%total + counts(i)
     s%sumOfSquares = s%sumOfSquares + counts(i)**2
     s%sigma(1) = s%sigma(1) + xpos(i) * counts(i)
     s%sigma(2) = s%sigma(2) + ypos(i) * counts(i)

     if (counts(i) < s%minval ) then
        s%minval = counts(i)
        s%minpos = (/xpos(i),ypos(i)/)
     endif
     if (counts(i) > s%maxval ) then
        s%maxval = counts(i)
        s%maxpos = (/xpos(i), ypos(i)/)
     endif

     if (xpos(i) < s%xrange(1) ) s%xrange(1) = xpos(i)
     if (xpos(i) > s%xrange(2) ) s%xrange(2) = xpos(i)
     if (ypos(i) < s%yrange(1) ) s%yrange(1) = ypos(i)
     if (ypos(i) > s%yrange(2) ) s%yrange(2) = ypos(i)

  enddo

   ! low mem requires col release
   do i=1,3
     call release(col(i))
   enddo
   if (doflag) call release(col(4))

 end subroutine processSubTable

end module tableStat




